/* 	General table driven syntax analyzer

 Copyright (C) 1994 by Yasumasa Uyama
	All Rights Reserved.
This is C++ version of Coco.
Source files of Coco are written by
Peter Rechenberg,Hanspeter Mossenbock,in Modula-2.
Published by Carl Hanser Verlag,Munich,
"Ein Compiler-Generator fur Mikrocomputer"(2.Auflage),1988.
 (Copyright Permission, 6 July 1994.)

========================================
This is a parser module generateed by CC from an attributed grammar.
Before calling the procedure Parse from the main program, initialize
the scanner (<grammarname>lex.MOD).
---------------------------------------------------------------------*/
//DEFINITION MODULE -->modulename;
Boolean printinput;  // (*trace the input tokens read*)
Boolean printnodes;  // (*trace the G-code interpretation*)

void Parse(Boolean &correct);
//end -->modulename.
-->implementation
/* 	General table driven syntax analyzer
========================================
This is a parser module generateed by CC from an attributed grammar.
Before calling the procedure Parse from the main program, initialize
the scanner (<grammarname>lex.MOD).
---------------------------------------------------------------------*/
//( IMPREMANTATION MODULE -->modulename )
#include <stdio.h>
#include <stdlib.h>
#include <mem.h>
#include <conio.h>
#include <dos.h>
#include "cctypes.h"
#include "ccsyn.h"
//from Error
struct Errornode{
  char txt[26];
  unsigned int l;
  Errornode * next;
  };
void SyntaxError(Errornode *, unsigned int ,unsigned int );
//from fileio
extern File *con;
void WriteCard(File * , unsigned int, int);
void WriteLn(File *  );
void WriteString(File *,char s[]);
//from ccsem
void Semant(unsigned int sem);
//from cclex
void GetSy(void);
extern unsigned int typ;
extern unsigned int line;
extern unsigned int col;
// FROM -->semantic analyzer IMPORT Semant;
// FROM -->input module IMPORT GetSy, typ, at, line, col;
-->declarations
//CONST  /*G-code instructions*/
#define t 0
#define ta 1
#define nt 2
#define nta 3
#define nts 4
#define ntas 5
#define any 6
#define anya 7
#define eps 8
#define epsa 9
#define jmp 10
#define ret 11
#define errdistmin 2
#define lmaxs 50
#define eofsy 0

struct Pragma{
  unsigned int sem2;
  unsigned int sem3;
  };
Pragma Pragmalist[cmaxp-cmaxt+1];
struct SymbolnodeSYN{
  unsigned int startpc;
  Boolean del;
  unsigned int first[1+cmaxt/16];
  };
struct TAB{
  unsigned int header[8];
  unsigned char code[(cmaxcode/2+2)*2];
  SymbolnodeSYN ntsymbols[cmaxs-cmaxp];
  unsigned int epsset[cmaxeps][cmaxt/16+1];
  unsigned int anyset[cmaxany][cmaxt/16+1];
  unsigned int  nra[cmaxp+1];
  Pragma ps[cmaxp-cmaxt+1];
  unsigned int namep[cmaxnamep+1];
  char name[(cmaxname/2+2)*2];
  };
TAB *tab;
extern Boolean correct;
unsigned int pc;
unsigned int errdist;
unsigned int newlacts[cmaxt+1];
unsigned int newpc[cmaxt+1];
unsigned int s[lmaxs+1];
unsigned int olds[lmaxs+1];
unsigned int lacts;
void GetSymInstr(unsigned int pc,unsigned int &opcode,unsigned int &sy,
	 unsigned int &nextpc,unsigned int &altpc);
void RestoreStack(void);
void SaveStack(void);
unsigned int StackElem(unsigned int i);
void Triple(unsigned int altroot);

/*Match     Check if sy is mumber of the specified set
------------------------------------------------------*/
Boolean Match(unsigned int sy,unsigned int set[]){
  if(set[sy/16]&(0x0001<<(sy%16))){return True; }
  else {return False; }
  }/* Match() */

/*NextSym     Get next symbol
----------------------------------------*/
void NextSym(void){
  while(1){
    GetSy();
    if(printinput){
      WriteString(con,"$(in:"); WriteCard(con,typ,3);
      WriteString(con,") ");
      }
    if(printnodes){
      WriteCard(con,lacts,3);WriteString(con,"| ");
      }
    if(typ<=cmaxt){
      return ;
      }
    if(correct && (tab->ps[typ-cmaxt].sem2!=0)){
      Semant(tab->ps[typ-cmaxt].sem2);
      }
    if(correct && (tab->ps[typ-cmaxt].sem3!=0)){
      Semant(tab->ps[typ-cmaxt].sem3);
      }
    if(typ==eofsy){
      return;
      }
    }
  }/* NextSym() */

/*====================== ERRORS ======================*/

/*AdjustPc    Adjust pc to next symbol instruction
---------------------------------------------------*/
void AdjustPc(unsigned int *pc){

  if(*pc==0){
    return;
    }
  while(1){
    switch(tab->code[*pc]){
      case t:
      case ta:
      case nt:
      case nta:
      case nts:
      case ntas:
      case any:
      case anya:
      case eps:
      case epsa:
		return;
      case jmp:
		*pc=256*(unsigned int)(tab->code[*pc+1]) +(unsigned int)(tab->code[*pc+2]);
		break;
      case ret:
		(*pc)=0;
		return;
      default:
		(*pc)++;
		break;
      }
    }
  }/* AdjustPc() */

/*GiveName
----------------*/
void GiveName(Errornode *q,unsigned int sy){
  unsigned int p,j;

  p=tab->namep[sy]; j=0;
  while(j<25 && tab->name[p+j-1]!='\0'){
    q->txt[j]=tab->name[p+j-1]; j++;
    }
  q->l=j;

  }/* GiveName() */

/*Error    Report syntax error
-------------------------------*/
void Error(unsigned int &pc,unsigned int &altroot){
  Errornode *e,*e1,*h;
  unsigned int i,j;
  unsigned int opcode,sy,nextpc,altpc,pc1;

  correct=False;
  if(errdist>=errdistmin){
    h=(Errornode *)malloc(sizeof(Errornode));
    GiveName(h,typ);
    h->next=NULL; e1=h;
    pc1=altroot; AdjustPc(&pc1);
    while(pc1>0){
      GetSymInstr(pc1,opcode,sy,nextpc,altpc);
      if(opcode<any){
	e=(Errornode *)malloc(sizeof(Errornode));
	GiveName(e,sy);
	e1->next=e; e1=e; e->next=NULL;
	}
      pc1=altpc;
      }
    SyntaxError(h,line,col);
    Triple(altroot); SaveStack();
    if(printnodes){
      WriteString(con,"$  typ    newpc  newlacts$");
      for(i=0;i<=cmaxt;i++){
	if(newpc[i]!=0){
	  WriteCard(con,i,5); WriteCard(con,newpc[i],10);
	  WriteCard(con,newlacts[i],10); WriteLn(con);
	  }/*if*/
	}/*for*/
      }/*if*/
    else{
      RestoreStack();
      }
    }
  while(newpc[typ]==0){
    if(printnodes){
      WriteString(con,"$(skip:"); WriteCard(con,typ,0);
      WriteString(con,") ");
      }
    NextSym();
    }
  pc=newpc[typ]; altroot=pc; lacts=newlacts[typ]; errdist=0;
  }/* Error() */

/*Fill   Fill triple llist with alt-chain string at pc
--------------------------------------------------------*/
void Fill(unsigned int pc,unsigned int lacts){
  unsigned int i,opcode,sy,nextpc,altpc;
  unsigned int s[cmaxt/16+1];

  AdjustPc(&pc);
  while(pc!=0){
    GetSymInstr(pc,opcode,sy,nextpc,altpc);
    switch(opcode){
      case t:
      case ta:
		newpc[sy]=pc; newlacts[sy]=lacts;
		break;
      case nt:
      case nta:
      case nts:
      case ntas:
		memcpy(s,tab->ntsymbols[sy-cmaxp-1].first,(cmaxt/16+1));
		for(i=0;i<=cmaxt;i++){
			if(Match(i,s)){
				newpc[i]=pc; newlacts[i]=lacts;
				}
			}
		if(tab->ntsymbols[sy-cmaxp-1].del){
			Fill(nextpc,lacts);
			}
		break;
      case eps:
      case epsa:
		Fill(nextpc,lacts);
		break;
      default:
		break;
      }/*switch*/
    pc=altpc;
    }/*while*/
  }/* Fill() */

/*FillSucc   Fill triple list with succ. of alt-chain at pc
-----------------------------------------------------------*/
void FillSucc(unsigned int pc,unsigned int lacts){
  unsigned int opcode,sy,nextpc,altpc;

  AdjustPc(&pc);
  while(pc>0){
    GetSymInstr(pc,opcode,sy,nextpc,altpc);
    if(nextpc>0){
      Fill(nextpc,lacts);
      }
    pc=altpc;
    }
  }/* FillSucc() */

/*GetSymInstr   Get G-code instruction at address pc
----------------------------------------------------*/
void GetSymInstr(unsigned int pc,unsigned int &opcode,unsigned int &sy,
		 unsigned int &nextpc,unsigned int &altpc){

  opcode=tab->code[pc];
  if(opcode<=epsa && opcode!=any){
    sy=tab->code[pc+1];
    }
  else{
    sy=0;
    }
  switch(opcode){
    case t:
    case nt:
    case eps:
		nextpc=pc+2; altpc=0;
		break;
    case ta:
    case nta:
    case anya:
    case epsa:
		nextpc=pc+4; altpc=256*tab->code[pc+2]+tab->code[pc+3];
		break;
    case nts:
		nextpc=pc+3; altpc=0;
		break;
    case ntas:
		nextpc=pc+5; altpc=256*tab->code[pc+2]+tab->code[pc+3];
		break;
    case any:
		nextpc=pc+1; altpc=0;
		break;
    default:
		break;
    }
  AdjustPc(&nextpc); AdjustPc(&altpc);
  }/* GetSymInstr() */

/*Triple     Fill triple list
-------------------------------*/
void Triple(unsigned int altroot){
  unsigned int i;

  for(i=0;i<=cmaxt;i++){
    newpc[i]=0; newlacts[i]=0;
    }
  for(i=1;i<=lacts;i++){
    FillSucc(StackElem(i),i-1);
    Fill(StackElem(i),i-1);
    }
  FillSucc(altroot,lacts);
  Fill(altroot,lacts);
  }/* Triple() */

/*================= END ERRORS ===================*/

/*================= SYNTAXSTACK ===================*/

/*Pop
---------------*/
void Pop(unsigned int &loc){

  if(lacts>0){
	loc=s[lacts]; lacts--;
	}
  else{
	WriteString(con,"--- Parser stack underflow.$");
	getch(); /* HALT */
	}
  if(printnodes){
	WriteString(con,"  pop");
	}
  }/* Pop() */

/*Push
------------*/
void Push(unsigned int loc){

  if(lacts<lmaxs){
	lacts++; s[lacts]=loc;
	}
  else{
	WriteString(con,"--- Parser stack overflow.$");
	getch();  /*HALT*/
	}
  if(printnodes){
	WriteString(con," push");
	}
  }/* Push() */

/*RestoreStack
------------------------*/
void RestoreStack(void){
  memcpy(s,olds,sizeof(s));
  }/* RestoreStack() */

/*SaveStack
--------------*/
void SaveStack(void){
  memcpy(olds,s,sizeof(s));
  } /* Savestack() */

/*StackElem
--------------*/
unsigned int StackElem(unsigned int i){
  return s[i];
  }/* StackElem() */

/*================== END SYNTAXSTACK =================*/


/*TableContents   A dirty trick to initialize the grammar tables
------------------------------------------------------------------*/
unsigned int TABLE[]={
-->tables
//	END TableContents;

/*Parse      Proper syntax analyzer
-------------------------------------*/
void Parse(Boolean &corr){
  unsigned int altroot;
  Boolean mustread;
  unsigned int opcode;
  Boolean running;
  unsigned int sy,tmp,tmp1;
  int i,j,k;

  tab=(TAB *)malloc(sizeof(TAB));
  for(i=0;i<8;i++){
    *((unsigned int *)tab+i)=*(TABLE+i);
    }
  for(i=8;i<=8;i++){
    tmp=*(TABLE+i);
    *((unsigned int *)tab+i)=(tmp&0xff00);
    }
  for(i=9;i<(8+cmaxcode/2+1);i++){
    tmp=*(TABLE+i-1);
    tmp1=*(TABLE+i);
    *((unsigned int *)tab+i)=((tmp1&0xff00)|(tmp&0x00ff));
    }
  for(i=8+(cmaxcode/2)+cmaxcode%2;
	i<(8+(cmaxcode/2)+cmaxcode%2+5*(cmaxs-cmaxp)
	 +cmaxeps*3+cmaxany*3+cmaxp+1
	 +(cmaxp-cmaxt+1)*2+cmaxnamep+1);i++){
    *((unsigned int *)tab+i+1+(cmaxcode+1)%2)=*(TABLE+i);
    }
  for(i=(8+(cmaxcode/2)+cmaxcode%2+5*(cmaxs-cmaxp)
		 +cmaxeps*3+cmaxany*3+cmaxp+1
		 +(cmaxp-cmaxt+1)*2+cmaxnamep+1);
		i<(8+(cmaxcode/2)+cmaxcode%2+5*(cmaxs-cmaxp)
		 +cmaxeps*3+cmaxany*3+cmaxp+1
		 +(cmaxp-cmaxt+1)*2+cmaxnamep+1
			 +(cmaxname/2+2));
				 i++){
    tmp=*(TABLE+i);
    *((unsigned int *)tab+i+1+(cmaxcode+1)%2)=((tmp&0x00ff)<<8)|((tmp&0xff00)>>8);
    }

  pc=cstartpc; altroot=pc;
  line=1; col=1;
  correct=True; mustread=True; running=True;

  while(running){
    opcode=tab->code[pc];
    if(mustread && opcode<=epsa){
      NextSym(); mustread=False; errdist++; altroot=pc;
      }
    if(printnodes){
      WriteCard(con,pc,5);
      }
    pc++;
    switch(opcode){
      case t:
	if(typ==tab->code[pc]){
	  if(typ==eofsy){
	    running=False;
	    }
	  else{
	    pc++; mustread=True;
	    }
	  }
	else{
	  Error(pc,altroot);
	  }
	break;
      case ta:
	if(typ==tab->code[pc]){
	  pc=pc+3; mustread=True;
	  }
	else{
	  pc=256*tab->code[pc+1]+tab->code[pc+2];
	  }
	break;
      case nt:
      case nts:
	sy=tab->code[pc];
	if(Match(typ,(tab->ntsymbols[sy-cmaxp-1].first)) ||
		 tab->ntsymbols[sy-cmaxp-1].del){
	  if(opcode==nts){
	    pc++; Semant(tab->code[pc]);
	    }
	  Push(pc+1); pc=tab->ntsymbols[sy-cmaxp-1].startpc;
	  altroot=pc;
	  }
	else{
	  Error(pc,altroot);
	  }
	break;
      case nta:
      case ntas:
	sy=tab->code[pc];
	if(Match(typ,tab->ntsymbols[sy-cmaxp-1].first)){
	  pc=pc+3;
	  if(opcode==ntas){
	    Semant(tab->code[pc]); pc++;
	    }
	  Push(pc); pc=tab->ntsymbols[sy-cmaxp-1].startpc;
	  altroot=pc;
	  }
	else{
	  pc=tab->code[pc+1]*256+tab->code[pc+2];
	  }
	break;
      case any:
	mustread=True;
	break;
      case anya:
	if(Match(typ,tab->anyset[tab->code[pc]-1])){
	  pc=pc+3; mustread=True;
	  }
	else{
	  pc=tab->code[pc+1]*256+tab->code[pc+2];
	  }
	  break;
	case eps:
	  if(Match(typ,tab->epsset[tab->code[pc]-1])){
	    pc++;
	    }
	  else{
	    Error(pc,altroot);
	    }
	  break;
	case epsa:
	  if(Match(typ,tab->epsset[tab->code[pc]-1])){
		pc=pc+3;
		}
	  else{
	    pc=tab->code[pc+1]*256+tab->code[pc+2];
	    }
	  break;
	case jmp:
		pc=tab->code[pc]*256+tab->code[pc+1];
		break;
	case ret:
		Pop(pc); altroot=pc;
		break;
	default:
	  if(correct){
	    Semant(opcode);
	    }
	  break;
	}
      }
    corr=correct;
  }/* Parse() */

/*ccsyn
----------------*/
void cocosyn(void){
  printinput=False;
  printnodes=False;
  errdist=100;
  lacts=0;
  }/* ccsyn() */
//END -->modulename
